/*
Copyright © 2025 filifa

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package markov

import (
	"errors"
	"math"
	"strconv"

	"gonum.org/v1/gonum/graph"
	"gonum.org/v1/gonum/graph/encoding"
	"gonum.org/v1/gonum/graph/multi"
	"gonum.org/v1/gonum/graph/simple"
	"gonum.org/v1/gonum/graph/topo"
	"gonum.org/v1/gonum/mat"
)

// AbsorbingMarkovChain is a graph representing an absorbing Markov chain.
type AbsorbingMarkovChain struct {
	*multi.WeightedDirectedGraph
}

// NewAbsorbingMarkovChain returns an absorbing Markov chain with no nodes or
// edges.
func NewAbsorbingMarkovChain() *AbsorbingMarkovChain {
	return &AbsorbingMarkovChain{WeightedDirectedGraph: multi.NewWeightedDirectedGraph()}
}

// IsValid returns true if the graph is a valid Markov chain.
func (g *AbsorbingMarkovChain) IsValid() bool {
	for nodes := g.Nodes(); nodes.Next(); {
		u := nodes.Node()
		if g.outWeightSum(u) != 1 {
			return false
		}
	}

	return true
}

func (g *AbsorbingMarkovChain) outWeightSum(u graph.Node) float64 {
	sum := 0.0
	for nodes := g.From(u.ID()); nodes.Next(); {
		v := nodes.Node()
		e := g.WeightedEdge(u.ID(), v.ID())
		if e != nil {
			sum += e.Weight()
		}
	}

	return sum
}

// AbsorbingNodes returns all the nodes in the Markov chain identified as
// absorbing nodes.
func (g *AbsorbingMarkovChain) AbsorbingNodes() []graph.Node {
	absorbingNodes := make([]graph.Node, 0)
	for nodes := g.Nodes(); nodes.Next(); {
		u := nodes.Node()
		successors := g.From(u.ID())
		if successors.Len() != 1 {
			continue
		}

		successors.Next()
		v := successors.Node()
		if u == v {
			absorbingNodes = append(absorbingNodes, u)
		}
	}

	return absorbingNodes
}

// IsAbsorbing returns true if the Markov chain is an absorbing chain. This
// means at least one node is an absorbing node and that every node can reach
// an absorbing node.
func (g *AbsorbingMarkovChain) IsAbsorbing() bool {
	absorbingNodes := g.AbsorbingNodes()
	if len(absorbingNodes) == 0 {
		return false
	}

	for nodes := g.Nodes(); nodes.Next(); {
		u := nodes.Node()
		if !g.canBeAbsorbed(u, absorbingNodes) {
			return false
		}
	}

	return true
}

func (g *AbsorbingMarkovChain) canBeAbsorbed(u graph.Node, absorbingNodes []graph.Node) bool {
	for _, v := range absorbingNodes {
		if topo.PathExistsIn(g, u, v) {
			return true
		}
	}

	return false
}

// AdjacencyMatrix returns the graph's adjacency matrix.
func (g *AbsorbingMarkovChain) AdjacencyMatrix() *mat.Dense {
	adj := simple.NewDirectedMatrix(g.Nodes().Len(), 0, 0, 0)
	for edges := g.WeightedEdges(); edges.Next(); {
		e := edges.WeightedEdge()
		if e.From() == e.To() {
			continue
		}

		adj.SetWeightedEdge(e)
	}

	a := mat.DenseCopyOf(adj.Matrix())

	nodes := adj.Nodes()
	for i := 0; nodes.Next(); i++ {
		id := nodes.Node().ID()
		u := g.Node(id)

		e := g.WeightedEdge(u.ID(), u.ID())
		if e != nil {
			a.Set(i, i, e.Weight())
		}
	}

	return a
}

// NewEdge returns a weightedEdge that can be added to the Markov chain.
func (g *AbsorbingMarkovChain) NewEdge(from, to graph.Node) graph.Edge {
	e := g.WeightedDirectedGraph.NewWeightedLine(from, to, math.NaN()).(multi.WeightedLine)
	return &weightedEdge{WeightedLine: e}
}

// NewNode returns a Node that can be added to the Markov chain.
func (g *AbsorbingMarkovChain) NewNode() graph.Node {
	return &Node{Node: g.WeightedDirectedGraph.NewNode()}
}

// SetEdge adds a weighted edge to the Markov chain.
func (g *AbsorbingMarkovChain) SetEdge(e graph.Edge) {
	g.WeightedDirectedGraph.SetWeightedLine(e.(*weightedEdge))
}

// weightedEdge is a DOT-aware multi.WeightedLine. By being a
// multi.WeightedLine, it allows for self-loops, which are important for
// absorbing Markov chains.
// TODO: this is a little confusing, maybe just have checks in the code that
// there's only one line in each edge?
type weightedEdge struct {
	multi.WeightedLine
}

// ReversedEdge returns a new weightedEdge with the same weight, but the
// direction reversed. It exists mainly to satisfy the graph.Edge interface.
func (e *weightedEdge) ReversedEdge() graph.Edge {
	revLine := multi.WeightedLine{F: e.T, T: e.F, W: e.W}
	return &weightedEdge{WeightedLine: revLine}
}

// SetAttribute enables storing the weight read from a DOT file. It errors if
// an attribute is read that can't be stored in a weightedEdge.
func (e *weightedEdge) SetAttribute(attr encoding.Attribute) error {
	var err error

	switch attr.Key {
	case "weight":
		e.W, err = strconv.ParseFloat(attr.Value, 64)
	default:
		err = errors.New("unknown key:" + attr.Key)
	}

	return err
}

// Node is a DOT-aware graph.Node.
type Node struct {
	graph.Node
	dotID string
}

// SetDOTID sets the node's DOT ID. It enables storing the node name read from
// a DOT file.
func (n *Node) SetDOTID(id string) {
	n.dotID = id
}

// DOTID returns the node's DOT ID.
func (n *Node) DOTID() string {
	return n.dotID
}
